@@ -10,17 +10,20 @@
 # License for the specific language governing permissions and limitations
 # under the License.
 
+from unittest import mock
+
 from keystoneauth1 import adapter
-import mock
 
 from openstack.baremetal.v1 import _common
 from openstack.baremetal.v1 import node
 from openstack import exceptions
 from openstack import resource
 from openstack.tests.unit import base
+from openstack import utils
 
 # NOTE: Sample data from api-ref doc
 FAKE = {
+    "boot_mode": "uefi",
     "chassis_uuid": "1",  # NOTE: missed in api-ref sample
     "clean_step": {},
     "console_enabled": False,
@@ -79,6 +82,7 @@ FAKE = {
     "raid_config": {},
     "reservation": None,
     "resource_class": None,
+    "secure_boot": True,
     "states": [
         {
             "href": "http://127.0.0.1:6385/v1/nodes/<NODE_ID>/states",
@@ -117,6 +121,7 @@ class TestNode(base.TestCase):
         self.assertEqual(FAKE['uuid'], sot.id)
         self.assertEqual(FAKE['name'], sot.name)
 
+        self.assertEqual(FAKE['boot_mode'], sot.boot_mode)
         self.assertEqual(FAKE['chassis_uuid'], sot.chassis_id)
         self.assertEqual(FAKE['clean_step'], sot.clean_step)
         self.assertEqual(FAKE['created_at'], sot.created_at)
@@ -143,6 +148,7 @@ class TestNode(base.TestCase):
         self.assertEqual(FAKE['raid_config'], sot.raid_config)
         self.assertEqual(FAKE['reservation'], sot.reservation)
         self.assertEqual(FAKE['resource_class'], sot.resource_class)
+        self.assertEqual(FAKE['secure_boot'], sot.is_secure_boot)
         self.assertEqual(FAKE['states'], sot.states)
         self.assertEqual(FAKE['target_provision_state'],
                          sot.target_provision_state)
@@ -294,6 +300,34 @@ class TestNodeSetProvisionState(base.TestCase):
                 headers=mock.ANY, microversion='1.56',
                 retriable_status_codes=_common.RETRIABLE_STATUS_CODES)
 
+    def test_deploy_with_deploy_steps(self):
+        deploy_steps = [{'interface': 'deploy', 'step': 'upgrade_fw'}]
+        result = self.node.set_provision_state(
+            self.session, 'active',
+            deploy_steps=deploy_steps)
+
+        self.assertIs(result, self.node)
+        self.session.put.assert_called_once_with(
+            'nodes/%s/states/provision' % self.node.id,
+            json={'target': 'active', 'deploy_steps': deploy_steps},
+            headers=mock.ANY, microversion='1.69',
+            retriable_status_codes=_common.RETRIABLE_STATUS_CODES
+        )
+
+    def test_rebuild_with_deploy_steps(self):
+        deploy_steps = [{'interface': 'deploy', 'step': 'upgrade_fw'}]
+        result = self.node.set_provision_state(
+            self.session, 'rebuild',
+            deploy_steps=deploy_steps)
+
+        self.assertIs(result, self.node)
+        self.session.put.assert_called_once_with(
+            'nodes/%s/states/provision' % self.node.id,
+            json={'target': 'rebuild', 'deploy_steps': deploy_steps},
+            headers=mock.ANY, microversion='1.69',
+            retriable_status_codes=_common.RETRIABLE_STATUS_CODES
+        )
+
 
 @mock.patch.object(node.Node, '_translate_response', mock.Mock())
 @mock.patch.object(node.Node, '_get_session', lambda self, x: x)
@@ -313,6 +347,7 @@ class TestNodeCreate(base.TestCase):
         self.session.post.side_effect = _change_state
 
     def test_available_old_version(self, mock_prov):
+        self.node.provision_state = 'available'
         result = self.node.create(self.session)
         self.assertIs(result, self.node)
         self.session.post.assert_called_once_with(
@@ -322,24 +357,16 @@ class TestNodeCreate(base.TestCase):
         self.assertFalse(mock_prov.called)
 
     def test_available_new_version(self, mock_prov):
-        def _change_state(*args, **kwargs):
-            self.node.provision_state = 'manageable'
-
         self.session.default_microversion = '1.11'
         self.node.provision_state = 'available'
-        self.new_state = 'enroll'
-        mock_prov.side_effect = _change_state
 
         result = self.node.create(self.session)
         self.assertIs(result, self.node)
         self.session.post.assert_called_once_with(
             mock.ANY, json={'driver': FAKE['driver']},
-            headers=mock.ANY, microversion=self.session.default_microversion,
+            headers=mock.ANY, microversion='1.10',
             params={})
-        mock_prov.assert_has_calls([
-            mock.call(self.node, self.session, 'manage', wait=True),
-            mock.call(self.node, self.session, 'provide', wait=True)
-        ])
+        mock_prov.assert_not_called()
 
     def test_no_enroll_in_old_version(self, mock_prov):
         self.node.provision_state = 'enroll'
@@ -563,6 +590,35 @@ class TestNodeWaitForReservation(base.TestCase):
         mock_fetch.assert_called_with(self.node, self.session)
 
 
+@mock.patch.object(exceptions, 'raise_from_response', mock.Mock())
+class TestNodeInjectNMI(base.TestCase):
+
+    def setUp(self):
+        super().setUp()
+        self.node = node.Node(**FAKE)
+        self.session = mock.Mock(spec=adapter.Adapter)
+        self.session.default_microversion = '1.29'
+        self.node = node.Node(**FAKE)
+
+    def test_inject_nmi(self):
+        self.node.inject_nmi(self.session)
+        self.session.put.assert_called_once_with(
+            'nodes/%s/management/inject_nmi' % FAKE['uuid'],
+            json={},
+            headers=mock.ANY,
+            microversion='1.29',
+            retriable_status_codes=_common.RETRIABLE_STATUS_CODES,
+        )
+
+    def test_incompatible_microversion(self):
+        self.session.default_microversion = '1.28'
+        self.assertRaises(
+            exceptions.NotSupported,
+            self.node.inject_nmi,
+            self.session,
+        )
+
+
 @mock.patch.object(node.Node, '_assert_microversion_for', _fake_assert)
 @mock.patch.object(exceptions, 'raise_from_response', mock.Mock())
 class TestNodeSetPowerState(base.TestCase):
@@ -713,15 +769,24 @@ class TestNodeMaintenance(base.TestCase):
 
 @mock.patch.object(node.Node, 'fetch', lambda self, session: self)
 @mock.patch.object(exceptions, 'raise_from_response', mock.Mock())
-class TestNodeSetBootDevice(base.TestCase):
+class TestNodeBootDevice(base.TestCase):
 
     def setUp(self):
-        super(TestNodeSetBootDevice, self).setUp()
+        super().setUp()
         self.node = node.Node(**FAKE)
         self.session = mock.Mock(spec=adapter.Adapter,
                                  default_microversion='1.1')
 
-    def test_node_set_boot_device(self):
+    def test_get_boot_device(self):
+        self.node.get_boot_device(self.session)
+        self.session.get.assert_called_once_with(
+            'nodes/%s/management/boot_device' % self.node.id,
+            headers=mock.ANY,
+            microversion=mock.ANY,
+            retriable_status_codes=_common.RETRIABLE_STATUS_CODES,
+        )
+
+    def test_set_boot_device(self):
         self.node.set_boot_device(self.session, 'pxe', persistent=False)
         self.session.put.assert_called_once_with(
             'nodes/%s/management/boot_device' % self.node.id,
@@ -729,7 +794,65 @@ class TestNodeSetBootDevice(base.TestCase):
             headers=mock.ANY, microversion=mock.ANY,
             retriable_status_codes=_common.RETRIABLE_STATUS_CODES)
 
+    def test_get_supported_boot_devices(self):
+        self.node.get_supported_boot_devices(self.session)
+        self.session.get.assert_called_once_with(
+            'nodes/%s/management/boot_device/supported' % self.node.id,
+            headers=mock.ANY,
+            microversion=mock.ANY,
+            retriable_status_codes=_common.RETRIABLE_STATUS_CODES,
+        )
 
+
+@mock.patch.object(utils, 'pick_microversion', lambda session, v: v)
+@mock.patch.object(node.Node, 'fetch', lambda self, session: self)
+@mock.patch.object(exceptions, 'raise_from_response', mock.Mock())
+class TestNodeSetBootMode(base.TestCase):
+
+    def setUp(self):
+        super(TestNodeSetBootMode, self).setUp()
+        self.node = node.Node(**FAKE)
+        self.session = mock.Mock(spec=adapter.Adapter,
+                                 default_microversion='1.1')
+
+    def test_node_set_boot_mode(self):
+        self.node.set_boot_mode(self.session, 'uefi')
+        self.session.put.assert_called_once_with(
+            'nodes/%s/states/boot_mode' % self.node.id,
+            json={'target': 'uefi'},
+            headers=mock.ANY, microversion=mock.ANY,
+            retriable_status_codes=_common.RETRIABLE_STATUS_CODES)
+
+    def test_node_set_boot_mode_invalid_mode(self):
+        self.assertRaises(ValueError,
+                          self.node.set_boot_mode, self.session, 'invalid-efi')
+
+
+@mock.patch.object(utils, 'pick_microversion', lambda session, v: v)
+@mock.patch.object(node.Node, 'fetch', lambda self, session: self)
+@mock.patch.object(exceptions, 'raise_from_response', mock.Mock())
+class TestNodeSetSecureBoot(base.TestCase):
+
+    def setUp(self):
+        super(TestNodeSetSecureBoot, self).setUp()
+        self.node = node.Node(**FAKE)
+        self.session = mock.Mock(spec=adapter.Adapter,
+                                 default_microversion='1.1')
+
+    def test_node_set_secure_boot(self):
+        self.node.set_secure_boot(self.session, True)
+        self.session.put.assert_called_once_with(
+            'nodes/%s/states/secure_boot' % self.node.id,
+            json={'target': True},
+            headers=mock.ANY, microversion=mock.ANY,
+            retriable_status_codes=_common.RETRIABLE_STATUS_CODES)
+
+    def test_node_set_secure_boot_invalid_none(self):
+        self.assertRaises(ValueError,
+                          self.node.set_secure_boot, self.session, None)
+
+
+@mock.patch.object(utils, 'pick_microversion', lambda session, v: v)
 @mock.patch.object(node.Node, 'fetch', lambda self, session: self)
 @mock.patch.object(exceptions, 'raise_from_response', mock.Mock())
 class TestNodeTraits(base.TestCase):
@@ -750,7 +873,8 @@ class TestNodeTraits(base.TestCase):
             retriable_status_codes=_common.RETRIABLE_STATUS_CODES)
 
     def test_remove_trait(self):
-        self.node.remove_trait(self.session, 'CUSTOM_FAKE')
+        self.assertTrue(self.node.remove_trait(self.session,
+                                               'CUSTOM_FAKE'))
         self.session.delete.assert_called_once_with(
             'nodes/%s/traits/%s' % (self.node.id, 'CUSTOM_FAKE'),
             headers=mock.ANY, microversion='1.37',
@@ -790,7 +914,7 @@ class TestNodePatch(base.TestCase):
         patch = {'path': 'test'}
         self.node.patch(self.session, patch=patch)
         mock_patch.assert_called_once()
-        kwargs = mock_patch.call_args.kwargs
+        kwargs = mock_patch.call_args[1]
         self.assertEqual(kwargs['patch'], {'path': 'test'})
 
     @mock.patch.object(resource.Resource, '_prepare_request', autospec=True)
@@ -801,12 +925,126 @@ class TestNodePatch(base.TestCase):
         self.node.patch(self.session, patch=patch, retry_on_conflict=True,
                         reset_interfaces=True)
         mock_prepreq.assert_called_once()
-        prepreq_kwargs = mock_prepreq.call_args.kwargs
+        prepreq_kwargs = mock_prepreq.call_args[1]
         self.assertEqual(prepreq_kwargs['params'],
                          [('reset_interfaces', True)])
         mock__commit.assert_called_once()
-        commit_args = mock__commit.call_args.args
-        commit_kwargs = mock__commit.call_args.kwargs
+        commit_args = mock__commit.call_args[0]
+        commit_kwargs = mock__commit.call_args[1]
         self.assertIn('1.45', commit_args)
         self.assertEqual(commit_kwargs['retry_on_conflict'], True)
         mock_patch.assert_not_called()
+
+
+@mock.patch('time.sleep', lambda _t: None)
+@mock.patch.object(node.Node, 'fetch', autospec=True)
+class TestNodeWaitForPowerState(base.TestCase):
+    def setUp(self):
+        super(TestNodeWaitForPowerState, self).setUp()
+        self.node = node.Node(**FAKE)
+        self.session = mock.Mock()
+
+    def test_success(self, mock_fetch):
+        self.node.power_state = 'power on'
+
+        def _get_side_effect(_self, session):
+            self.node.power_state = 'power off'
+            self.assertIs(session, self.session)
+
+        mock_fetch.side_effect = _get_side_effect
+
+        node = self.node.wait_for_power_state(self.session, 'power off')
+        self.assertIs(node, self.node)
+
+    def test_timeout(self, mock_fetch):
+        self.node.power_state = 'power on'
+        self.assertRaises(exceptions.ResourceTimeout,
+                          self.node.wait_for_power_state,
+                          self.session, 'power off', timeout=0.001)
+
+
+@mock.patch.object(utils, 'pick_microversion', lambda session, v: v)
+@mock.patch.object(node.Node, 'fetch', lambda self, session: self)
+@mock.patch.object(exceptions, 'raise_from_response', mock.Mock())
+class TestNodePassthru(object):
+    def setUp(self):
+        super(TestNodePassthru, self).setUp()
+        self.node = node.Node(**FAKE)
+        self.session = node.Mock(spec=adapter.Adapter,
+                                 default_microversion='1.37')
+        self.session.log = mock.Mock()
+
+    def test_get_passthru(self):
+        self.node.call_vendor_passthru(self.session, "GET", "test_method")
+        self.session.get.assert_called_once_with(
+            'nodes/%s/vendor_passthru?method=test_method' % self.node.id,
+            headers=mock.ANY, microversion='1.37',
+            retriable_status_codes=_common.RETRIABLE_STATUS_CODES)
+
+    def test_post_passthru(self):
+        self.node.call_vendor_passthru(self.session, "POST", "test_method")
+        self.session.post.assert_called_once_with(
+            'nodes/%s/vendor_passthru?method=test_method' % self.node.id,
+            headers=mock.ANY, microversion='1.37',
+            retriable_status_codes=_common.RETRIABLE_STATUS_CODES)
+
+    def test_put_passthru(self):
+        self.node.call_vendor_passthru(self.session, "PUT", "test_method")
+        self.session.put.assert_called_once_with(
+            'nodes/%s/vendor_passthru?method=test_method' % self.node.id,
+            headers=mock.ANY, microversion='1.37',
+            retriable_status_codes=_common.RETRIABLE_STATUS_CODES)
+
+    def test_delete_passthru(self):
+        self.node.call_vendor_passthru(self.session, "DELETE", "test_method")
+        self.session.delete.assert_called_once_with(
+            'nodes/%s/vendor_passthru?method=test_method' % self.node.id,
+            headers=mock.ANY, microversion='1.37',
+            retriable_status_codes=_common.RETRIABLE_STATUS_CODES)
+
+    def test_list_passthru(self):
+        self.node.list_vendor_passthru(self.session)
+        self.session.get.assert_called_once_with(
+            'nodes/%s/vendor_passthru/methods' % self.node.id,
+            headers=mock.ANY, microversion='1.37',
+            retriable_status_codes=_common.RETRIABLE_STATUS_CODES)
+
+
+@mock.patch.object(node.Node, 'fetch', lambda self, session: self)
+@mock.patch.object(exceptions, 'raise_from_response', mock.Mock())
+class TestNodeConsole(base.TestCase):
+
+    def setUp(self):
+        super().setUp()
+        self.node = node.Node(**FAKE)
+        self.session = mock.Mock(
+            spec=adapter.Adapter,
+            default_microversion='1.1',
+        )
+
+    def test_get_console(self):
+        self.node.get_console(self.session)
+        self.session.get.assert_called_once_with(
+            'nodes/%s/states/console' % self.node.id,
+            headers=mock.ANY,
+            microversion=mock.ANY,
+            retriable_status_codes=_common.RETRIABLE_STATUS_CODES,
+        )
+
+    def test_set_console_mode(self):
+        self.node.set_console_mode(self.session, True)
+        self.session.put.assert_called_once_with(
+            'nodes/%s/states/console' % self.node.id,
+            json={'enabled': True},
+            headers=mock.ANY,
+            microversion=mock.ANY,
+            retriable_status_codes=_common.RETRIABLE_STATUS_CODES,
+        )
+
+    def test_set_console_mode_invalid_enabled(self):
+        self.assertRaises(
+            ValueError,
+            self.node.set_console_mode,
+            self.session,
+            'true',  # not a bool
+        )
